from django.conf import settings
from rest_framework.settings import APISettings


DEFAULTS = {
    "USER_IDENTITY_FIELDS": ("email",),
    "REQUIRED_AUTH_FIELDS": (),
    "OPTIONAL_AUTH_FIELDS": ("otp", "password"),
    "TEST_USER_EMAIL": None,
    "EMAIL_CONFIRMED_FIELD": "email_confirmed",
    "OTP_EMAIL_UPDATE": True,
}

IMPORT_STRINGS = []

api_settings = APISettings(getattr(settings, "DF_AUTH", IMPORT_STRINGS), DEFAULTS)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DFAuthConfig(AppConfig):
    default_auto_field = "hashid_field.BigHashidAutoField"
    api_path = "auth/"
    name = "df_auth"
    verbose_name = _("DjangoFlow Auth")

    def ready(self):
        try:
            import df_auth.signals  # noqa F401
        except ImportError:
            pass



from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django.contrib.auth.models import update_last_login
from django.utils.module_loading import import_string
from rest_framework import exceptions
from rest_framework import serializers
from rest_framework_simplejwt.settings import api_settings as simplejwt_settings
from social_core.exceptions import AuthCanceled, AuthForbidden
from social_django.utils import load_backend
from social_django.models import DjangoStorage
from ..settings import api_settings
from ..strategy import DRFStrategy

User = get_user_model()

AUTHENTICATION_BACKENDS = [
    import_string(backend) for backend in settings.AUTHENTICATION_BACKENDS
]


class AbstractIdentitySerializer(serializers.Serializer):
    user = None

    def get_fields(self):
        return super().get_fields() | {
            f: serializers.CharField(write_only=True, required=False)
            for f in api_settings.USER_IDENTITY_FIELDS
        }

    def validate_email(self, value):
        return User.objects.normalize_email(value)


class TokenCreateSerializer(serializers.Serializer):
    token = serializers.CharField(read_only=True)
    token_class = simplejwt_settings.AUTH_TOKEN_CLASSES[0]
    user = None

    @classmethod
    def get_token(cls, user):
        return cls.token_class.for_user(user)

    def validate(self, attrs):
        if not simplejwt_settings.USER_AUTHENTICATION_RULE(self.user):
            raise exceptions.AuthenticationFailed()

        token = self.get_token(self.user)

        attrs["token"] = str(token)

        if simplejwt_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, self.user)

        return attrs


class TokenObtainSerializer(AbstractIdentitySerializer, TokenCreateSerializer):
    def validate(self, attrs):
        """
        Remove empty values to pass to authenticate or send_otp
        """
        attrs = {k: v for k, v in attrs.items() if v}
        self.user = authenticate(**attrs, **self.context)
        return super().validate(attrs)

    def get_fields(self):
        fields = super().get_fields()
        if api_settings.REQUIRED_AUTH_FIELDS:
            fields.update(
                {
                    f: serializers.CharField(write_only=True, required=True)
                    for f in api_settings.REQUIRED_AUTH_FIELDS
                }
            )

        if api_settings.OPTIONAL_AUTH_FIELDS:
            fields.update(
                {
                    f: serializers.CharField(write_only=True, required=False)
                    for f in api_settings.OPTIONAL_AUTH_FIELDS
                }
            )
        return fields


class OTPObtainSerializer(AbstractIdentitySerializer):
    def get_fields(self):
        return super().get_fields() | {
            f: serializers.CharField(write_only=True, required=False)
            for f in zip(
                api_settings.REQUIRED_AUTH_FIELDS, api_settings.OPTIONAL_AUTH_FIELDS
            )
        }

    def validate(self, attrs):
        attrs = {k: v for k, v in attrs.items() if v}
        attrs = super().validate(attrs)
        for backend in AUTHENTICATION_BACKENDS:
            if hasattr(backend, "generate_challenge"):
                backend().generate_challenge(**attrs, **self.context)
        return attrs


class SocialTokenObtainSerializer(TokenCreateSerializer):
    access_token = serializers.CharField(write_only=True)
    provider = serializers.ChoiceField(
        choices=[
            (backend.name, backend.name)
            for backend in AUTHENTICATION_BACKENDS
            if hasattr(backend, "name")
        ],
        write_only=True,
    )

    response = serializers.JSONField(read_only=True)

    def validate(self, attrs):
        request = self.context["request"]
        request.social_strategy = DRFStrategy(DjangoStorage, request)
        request.backend = load_backend(request.social_strategy, attrs["provider"], redirect_uri=None)

        try:
            self.user = request.backend.do_auth(attrs['access_token'])
        except (AuthCanceled, AuthForbidden):
            raise exceptions.AuthenticationFailed()

        return super().validate(attrs)



DF_AUTH_APPS = [
    "df_auth",
    "rest_framework_simplejwt",
    "django_otp",
    "django_otp.plugins.otp_email",
    "django_otp.plugins.otp_totp",
    "django_otp.plugins.otp_static",
    "otp_twilio",
    "social_django",
]



from social_django.strategy import DjangoStrategy


class DRFStrategy(DjangoStrategy):
    def __init__(self, storage, request):
        self.request = request
        super().__init__(storage, request)

    def request_data(self):
        return self.request.data



from .settings import api_settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django_otp.plugins.otp_email.models import EmailDevice


User = get_user_model()


class EmailOTPBackend(ModelBackend):
    @staticmethod
    def get_users(email):
        return User.objects.filter(is_active=True, email=email)

    def authenticate(self, request=None, email=None, otp=None, **kwargs):
        if email and otp:
            for user in self.get_users(email):
                if user.email == api_settings.TEST_USER_EMAIL:
                    return user
                if self.user_can_authenticate(user):
                    devices = EmailDevice.objects.filter(user=user)

                    if devices:
                        for device in devices.filter(confirmed=True):
                            if device.verify_token(otp):
                                updated_fields = []
                                if not getattr(
                                    user, api_settings.EMAIL_CONFIRMED_FIELD, True
                                ):
                                    setattr(user, api_settings.EMAIL_CONFIRMED_FIELD, True)
                                    updated_fields.append(
                                        api_settings.EMAIL_CONFIRMED_FIELD
                                    )

                                if (
                                    api_settings.OTP_EMAIL_UPDATE
                                    and device.email
                                    and user.email != device.email
                                ):
                                    user.email = device.email
                                    updated_fields.append("email")

                                if updated_fields:
                                    user.save(update_fields=updated_fields)

                                return user
                            raise ValidationError(
                                _("Wrong or expired one-time password")
                            )

    def generate_challenge(self, request=None, user=None, email=None, **kwargs):
        users = [user] if user else self.get_users(email)
        if email:
            for user in users:
                device = EmailDevice.objects.get_or_create(user=user, email=email)[0]
                device.generate_challenge()



"""Djangoflow URL Configuration

Add these to your root URLconf:
    urlpatterns = [
        ...
        path('auth/', include('df_auth.urls'))
    ]

"""
from rest_framework.routers import DefaultRouter

from .viewsets import OTPViewSet
from .viewsets import SocialTokenViewSet
from .viewsets import TokenViewSet

router = DefaultRouter()
router.register("token", TokenViewSet, basename="token")
router.register("otp", OTPViewSet, basename="otp")
router.register("social", SocialTokenViewSet, basename="social")

urlpatterns = router.urls



from .serializers import OTPObtainSerializer
from .serializers import SocialTokenObtainSerializer
from .serializers import TokenObtainSerializer
from django.conf import settings
from rest_framework import permissions
from rest_framework import response
from rest_framework import status
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.settings import import_string
from rest_framework_simplejwt.settings import api_settings as simple_jwt_settings


class ValidationOnlyCreateViewSet(viewsets.GenericViewSet):
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        return response.Response(serializer.data, status=status.HTTP_200_OK)


class TokenViewSet(ValidationOnlyCreateViewSet):
    serializer_class = TokenObtainSerializer
    permission_classes = (permissions.AllowAny,)

    @action(
        methods=["post"],
        detail=False,
        serializer_class=import_string(simple_jwt_settings.TOKEN_REFRESH_SERIALIZER),
    )
    def refresh(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    @action(
        methods=["post"],
        detail=False,
        serializer_class=import_string(simple_jwt_settings.TOKEN_VERIFY_SERIALIZER),
    )
    def verify(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    @action(
        methods=["post"],
        detail=False,
        serializer_class=import_string(simple_jwt_settings.TOKEN_BLACKLIST_SERIALIZER),
    )
    def blacklist(self, request, *args, **kwargs):
        if "rest_framework_simplejwt.token_blacklist" not in settings.INSTALLED_APPS:
            raise NotImplementedError

        return self.create(request, *args, **kwargs)


class OTPViewSet(ValidationOnlyCreateViewSet):
    serializer_class = OTPObtainSerializer
    permission_classes = (permissions.AllowAny,)


class SocialTokenViewSet(ValidationOnlyCreateViewSet):
    serializer_class = SocialTokenObtainSerializer
    permission_classes = (permissions.AllowAny,)
